//! This crate has procedural macros for compiling shader source into SPIR-V
//! bytes at compile time.
//!
//! # Nightly Only
//!
//! Requires the use of the `proc_macro_hygiene` feature because [unfortunately
//! 1.30 was a lie](https://github.com/rust-lang/blog.rust-lang.org/issues/285).

// Note(Lokathor): this extern crate is necessary even in 2018 for whatever
// reason that I'm sure is stupid.
extern crate proc_macro;

use core::str::FromStr;
use proc_macro::TokenStream;
use shaderc::ShaderKind;
use std::{
  fs::File,
  io::{prelude::*, BufReader},
};
use syn::{
  parse::{Parse, ParseStream, Result},
  parse_macro_input, Error, Ident, LitStr, Token,
};

struct SPIRVFromFile {
  kind: ShaderKind,
  filename: LitStr,
  entry_point: LitStr,
}

impl Parse for SPIRVFromFile {
  fn parse(input: ParseStream) -> Result<Self> {
    let kind_ident = input.parse::<Ident>()?;
    let kind_string = kind_ident.to_string();
    let kind = match kind_string.as_ref() {
      "Vertex" => ShaderKind::Vertex,
      "Fragment" => ShaderKind::Fragment,
      "Compute" => ShaderKind::Compute,
      "Geometry" => ShaderKind::Geometry,
      "TessControl" => ShaderKind::TessControl,
      "TessEvaluation" => ShaderKind::TessEvaluation,
      "InferFromSource" => ShaderKind::InferFromSource,
      "DefaultVertex" => ShaderKind::DefaultVertex,
      "DefaultFragment" => ShaderKind::DefaultFragment,
      "DefaultCompute" => ShaderKind::DefaultCompute,
      "DefaultGeometry" => ShaderKind::DefaultGeometry,
      "DefaultTessControl" => ShaderKind::DefaultTessControl,
      "DefaultTessEvaluation" => ShaderKind::DefaultTessEvaluation,
      "SpirvAssembly" => ShaderKind::SpirvAssembly,
      _ => {
        return Err(Error::new(
          kind_ident.span(),
          format!("Unknown ShaderKind value: {:?}", kind_string),
        ));
      }
    };
    let _ = input.parse::<Token![,]>()?;
    let filename = input.parse::<LitStr>()?;
    let _ = input.parse::<Token![,]>()?;
    let entry_point = input.parse::<LitStr>()?;
    Ok(SPIRVFromFile {
      kind,
      filename,
      entry_point,
    })
  }
}

/// `spirv_from_file!(ShaderKind, "file_path", "entry_point")`
///
/// * [ShaderKind](shaderc::ShaderKind) is just the final variant bit of the
///   enum from the `shaderc` crate. So use `Fragment`, not
///   `ShaderKind::Fragment` or anything else. Technically this just parses an
///   Ident, so you actually don't even need the `ShaderKind` type to be in
///   scope at all, you just write the name of one of those variants.
/// * String literal containing the path to the shader source. This is passed
///   directly to [File::open](std::fs::File::open). The "present working
///   directory" for proc macros is the crate root, so relative paths should
///   start from there.
/// * String literal for the name of the entry point of the shader module. By
///   convention this is is usually "main", but do what you want.
///
/// Success produces a `&[u8]` TokenStream. You probably want to assign this
/// into a `const` somewhere in your program.
///
/// Like this:
///
/// ```
/// #![feature(proc_macro_hygiene)]
///
/// use proc_spirv::spirv_from_file;
///
/// const FRAG_SPIRV: &[u8] = spirv_from_file!(Fragment, "tests/frag_glsl.frag", "main");
/// ```
#[proc_macro]
pub fn spirv_from_file(input: TokenStream) -> TokenStream {
  let SPIRVFromFile {
    kind,
    filename,
    entry_point,
  } = parse_macro_input!(input as SPIRVFromFile);

  let filename_string = filename.value();
  let file_handle = File::open(&filename_string)
    .unwrap_or_else(|e| panic!("Failed to open the file specified: {:?}", e));
  let mut buf_reader = BufReader::new(file_handle);
  let mut file_contents = String::new();
  buf_reader
    .read_to_string(&mut file_contents)
    .unwrap_or_else(|e| panic!("Failed to read the file contents: {:?}", e));

  let entry_point_string = entry_point.value();

  let mut compiler = shaderc::Compiler::new().expect("Could not initialize the shaderc compiler!");
  //
  let source_text: &str = &file_contents;
  let shader_kind: ShaderKind = kind;
  let input_file_name: &str = &filename_string;
  let entry_point_name: &str = &entry_point_string;
  let additional_options: Option<&shaderc::CompileOptions> = None;
  //
  let compile_result = compiler.compile_into_spirv(
    source_text,
    shader_kind,
    input_file_name,
    entry_point_name,
    additional_options,
  );

  let artifact = compile_result.unwrap_or_else(|e| panic!("SPIR-V compilation error: {}", e));

  let out_string = format!("&{:?}", artifact.as_binary_u8());

  TokenStream::from_str(&out_string).unwrap_or_else(|e| panic!("SPIR-V compilation error: {:?}", e))
}
